/*
 * Copyright (c) 2005-2007 CodeDroids ApS (http://www.codedroids.com)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * For further information about CodeDroids ApS, please see the
 * company website: http://www.codedroids.com
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * $LastChangedRevision: 182 $
 * $LastChangedDate: 2007-07-05 11:13:28 +0200 (Thu, 05 Jul 2007) $
 * 
 */

package com.codedroids.jam;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.NamedNodeMap;
import org.xml.sax.SAXException;

import com.codedroids.jam.JspAppNode;
import com.codedroids.jam.JspAppState;

public class JspApp implements JspAppNode {

	public static final String FILE_EXT = ".jam.xml";

	/**
	 * Returns an instance from an already parsed XML file
	 * @param xmlfile
	 * @return
	 */
	public static JspApp getInstance(Element root) throws ParserConfigurationException, SAXException, IOException
	{
		JspApp app = new JspApp();
		app.parseNode(root);
		return app;
	}
	
	/**
	 * Collect text by concatenating all TEXT and CDATA-nodes into string
	 * @param node
	 * @return String containing the text or null
	 */
	public static String collectText(Node node) {
		Node child = node.getFirstChild();
		StringBuffer retval = new StringBuffer();
		while(child != null) {
			switch(child.getNodeType()) {
				case Node.TEXT_NODE:
				case Node.CDATA_SECTION_NODE:
					retval.append(child.getNodeValue());
			}
			child = child.getNextSibling();
		}
		if(retval.length()==0)
			return null;
		else
			return retval.toString();
	}
	
	//
	// Instance
	//
	
	private String name; // application name
	private String version; // format version
	private String initialState; // start state (defaults to the first state if not given)
	private String prefix; // prefix used for attributes and parameters, defaults to "jam:"
	private List states;
	
	// 
	// Constructors
	//
	
	public JspApp() {
		super();
	}

	
	//
	// Accessors
	//
	
	public String getName() {
		return name;
	}

	public void setName(String str) {
		this.name = str;
	}
	
	public String getInitialState() {
		return initialState;
	}

	public void setInitialState(String s) {
		this.initialState = s;
	}
	
	public String getPrefix() {
		if(prefix==null)
			return JAM_DEFAULT_PREFIX;
		else
			return prefix;
	}

	public void setPrefix(String str) {
		this.prefix = str;
	}
	
	public List getStates() {
		return states;
	}

	public void setStates(List states) {
		this.states = states;
	}
	
	public void addState(JspAppState s) {
		if(this.states == null)
			this.states = new ArrayList();
		this.states.add(s);
	}
	
	public void removeState(JspAppState s) {
		if(this.states != null)
			this.states.remove(s);
	}
	
	public String getVersion() {
		if(version==null)
			return "1";
		else
			return version;
 	}

	public void setVersion(String str) {
		this.version = str;
	}
	
	//
	// Convenience
	//
	
	public JspAppState stateWithName(String name) {
		JspAppState result = null;
		if(getStates() != null && getStates().size() > 0) {
			Iterator it = getStates().iterator();
			while(it.hasNext()) {
				JspAppState s = (JspAppState)it.next();
				if(s.getName().equals(name)) {
					result = s;
					break;
				}
			}
		}
		return result;
	}
	
	
	//
	// Stuff
	//
	
	public String toString() {
		return toXML("");
	}
	

	public String toXML(String indent) {
		StringBuffer b = new StringBuffer();
		b.append("\n"+indent+"<"+JAM_APP_TAG+" "+JAM_NAME_ATTR+"=\""+getName()+"\" "+JAM_VERS_ATTR+"=\""+getVersion()+"\">");
		b.append("\n"+indent+PP_INDENT+"<"+JAM_PREFIX_TAG+">" + getPrefix() + "</"+JAM_PREFIX_TAG+">");
		b.append("\n"+indent+PP_INDENT+"<"+JAM_INIT_TAG+">" + getInitialState() + "</"+JAM_INIT_TAG+">");
		if(getStates() != null) {
			b.append("\n"+indent+PP_INDENT+"<"+JAM_STATES_TAG+">");
			Iterator it = getStates().iterator();
			while(it.hasNext()){
				JspAppState f = (JspAppState)it.next();
				b.append(f.toXML(indent+PP_INDENT+PP_INDENT));
			}
			b.append("\n"+indent+PP_INDENT+"</"+JAM_STATES_TAG+">");
		}
		b.append("\n</"+indent+JAM_APP_TAG+">");
		return b.toString();
	}
	
    //
	// Parsing
	//
	
	public JspAppNode parseNode(Node root)
	{
		NamedNodeMap attrs = root.getAttributes();
		for(int i=0; i<attrs.getLength(); i++) {
			Node attr = attrs.item(i);
			String an = attr.getNodeName();
			if(an.equals(JAM_NAME_ATTR)) {
				setName(JspApp.collectText(attr));
			}
			else if(an.equals(JAM_VERS_ATTR)) {
				setVersion(JspApp.collectText(attr));
			}
		}
		Node child = root.getFirstChild();
		String initStateName = null;
		while( child != null) {
			short nt = child.getNodeType();
			switch(nt) {
				case Node.ELEMENT_NODE:
					String nn = child.getNodeName();
					if (nn.equals(JAM_PREFIX_TAG)) {
						setPrefix(JspApp.collectText(child));
					}
					else if (nn.equals(JAM_INIT_TAG)) {
						initStateName = JspApp.collectText(child);
					}
					else if (nn.equals(JAM_STATES_TAG)) {
						parseStates(child);
					}
					break;
				default:		
					break;
			}
			child = child.getNextSibling();
		}
		// Default init state to first state if not explicitly given
		if(getStates() != null && getStates().size() > 0) {
			if(initStateName == null) {
				setInitialState(((JspAppState)getStates().get(0)).getName());
			} else {
				setInitialState(initStateName);
			}
		}
		return this;
	}
	
	protected void parseStates(Node states)
	{
		NodeList nl = states.getChildNodes();
		int len = nl.getLength();
		for(int i = 0; i < len; i++) {
			Node state = nl.item(i);
			short ft = state.getNodeType();
			switch(ft) {
				case Node.ELEMENT_NODE:
					String fn = state.getNodeName();
					if (fn.equals(JAM_STATE_TAG)) {
						addState((JspAppState)new JspAppState(this).parseNode(state));
					}
					break;
				default:		
					break;
			}
		}
	}
}
