/*
 * Copyright (c) 2004-2007 CodeDroids ApS (http://www.codedroids.com)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * For further information about CodeDroids ApS, please see the
 * company website: http://www.codedroids.com
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * $LastChangedRevision: 168 $
 * $LastChangedDate: 2007-03-22 11:46:00 +0100 (Thu, 22 Mar 2007) $
 * 
 */

package com.codedroids.oc;

import java.io.*;
import java.net.MalformedURLException;
import java.util.HashMap;
import java.util.Map;
import java.util.Iterator;
import java.util.List;
import java.util.ArrayList;
import java.util.regex.*;
import java.util.Properties;

import javax.servlet.http.HttpServletRequest;
import javax.xml.parsers.*;
import org.w3c.dom.*;
import org.xml.sax.*;

import org.opencms.main.CmsException;
import org.opencms.main.OpenCms;
import org.opencms.site.CmsSite;
import org.opencms.site.CmsSiteManager;
import org.opencms.site.CmsSiteMatcher;
import org.opencms.staticexport.CmsLinkManager;
import org.opencms.file.CmsFile;
import org.opencms.file.CmsObject;
import org.opencms.file.CmsResource;
import org.opencms.i18n.CmsEncoder;
import org.opencms.i18n.CmsMessages;
import org.opencms.jsp.CmsJspBean;
import org.opencms.jsp.CmsJspActionElement;
import org.opencms.jsp.CmsJspNavBuilder;
import org.opencms.jsp.CmsJspNavElement;
import org.opencms.jsp.CmsJspTagProperty;
import org.opencms.file.CmsResourceFilter;
import com.codedroids.j.StringUtils;

public class OCUtils {
	
	/** The log object for this class. */
    //private static final Log LOG = CmsLog.getLog(CmsModule.class);
	
	/**
	 * 
	 */
	public OCUtils() {
		super();
	}
	
	/**
	 * Load a xml file and parse it - consider using the OCCachedXmlDocument since it caches the resulting dom.
	 * 
	 * @param dir
	 * @return DOM document.
	 */
	public static Document loadXMLFileFromVFS(CmsObject cmso, String file) throws ParserConfigurationException, SAXException, IOException, CmsException
	{
		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
		DocumentBuilder builder = factory.newDocumentBuilder();
		CmsFile f = cmso.readFile(file);
		ByteArrayInputStream is = new ByteArrayInputStream(f.getContents());
		return builder.parse(is);  
	}
			
	//
	// Methods for juggling module names and paths
	//
	
	public static final String MODULES_DIR = "/system/modules";
	
	public static String pathForModule(String module)
	{
		return MODULES_DIR + "/" + module + "/";
	}

	/**
	 * Returns "foo.bar" from following examples: "/system/modules/foo.bar" or "/system/modules/foo.bar/templates/foobar".
	 * 
	 * @param path
	 * @return
	 */
	public static String moduleNameFromPath(String path) {
		if(path != null) {
			int ndx = path.indexOf(OCUtils.MODULES_DIR);
			int len = path.length(); 
			int start = OCUtils.MODULES_DIR.length() + 1; // to skip slash
			if(ndx==0 && len > start) {
				int end = path.indexOf("/", start); 
				if(end == -1)
					end = len;
				if(end > start) {
					return path.substring(start, end); 
				}
			}
		}
		return null;
	}
	
	/**
	 * Check the element uri to see if the element is inside a module and if so returns the name of that module.
	 * 
	 * @param cms
	 * @return Name of a module or null if the element uri is not inside a module
	 */
	public static String moduleName(CmsJspBean cms) {
		org.opencms.flex.CmsFlexController controller = org.opencms.flex.CmsFlexController.getController(cms.getRequest());
		return  moduleNameFromPath(controller.getCurrentRequest().getElementUri());
	}

	/**
	 * List modules by scanning the /system/modules dir. Note that this is not exact in terms of what a module is
	 * in OC 6 - use {@link CmsModuleManager} instead.
	 * 
	 * @param cmso Initialized CmsObject
	 * @return A (possibly empty) list of module names
	 */
	public static List moduleNames(CmsObject cmso) throws CmsException
	{
		List names = new ArrayList();
		List allModules = cmso.getResourcesInFolder(MODULES_DIR, CmsResourceFilter.ALL);
		Iterator mods = allModules.iterator();
		while(mods.hasNext()) {
			String name = ((CmsResource)mods.next()).getName();
			names.add(name);
		}
		return names;
	}
	
	//
	// Convenience methods for dealing with setup like properties files
	//
	
	/**
	 * Returns a cached (and localized) Properties object which serves as the site setup.
	 * 
	 * <p>Searched for the <code>Setup</code> property value for the current resource. The value must specify a properties file
	 * relative to the current site root. If the property is not set, then an attempt is made to locate the "/site.properties" file.
	 * <p>The path is used to lookup the properties in the default application OCCache or to load it into the cache if not already there.
	 * 
	 * @param cms
	 * @return
	 * @throws CmsException
	 */
	public static Properties siteSetup(CmsJspActionElement cms) throws CmsException {
		String path = cms.property("Setup", "search.uri", null);
		if(path==null)
			path="/site.properties";
		OCCache ca = OCCache.getAppInstance(cms);
		return OCCachedProperties.getCachedProperties(ca, cms, path, true);
	}

	/**
	 * Returns a cached (and localized) Properties object if the name is given or if a module can be identified from the request
	 * and the module includes a <code>.../setup/module.properties</code> file (and possibly language variants).
	 * 
	 * <p>If <code>moduleName()</code> returns a name, this is combined with <code>pathForModule(name) + setup/module.properties</code>
	 * to define the path for the setup file. If the module path cannot be identified the method returns null.
	 * <p>The path is used to lookup the properties in the default application OCCache or to load it into the cache if not already there.
	 * 
	 * @param cms
	 * @param name Optional module name, if null the name is found from the current request 
	 * @return
	 * @throws CmsException
	 */
	public static Properties moduleSetup(CmsJspBean cms, String name) throws CmsException {
		if(name == null) {
			if((name = moduleName(cms))==null)
				return null; // cannot identify the module from current request, bail out
		}
		OCCache ca = OCCache.getAppInstance(cms);
		return OCCachedProperties.getCachedProperties(ca, cms, pathForModule(name) + "setup/module.properties", true);
	}
	
	/**
	 * Convenience for calling {@link String property(String name, String file, String defaultValue, boolean escapeHtml, CmsJspActionElement cms, Properties site, Properties module)}
	 * - looks only at the current resource, the returned value is not html escaped.
	 * @param name
	 * @param cms
	 * @param site
	 * @param module
	 * @return
	 */
	public String property(String name, CmsJspActionElement cms, Properties site, Properties module) {
		return property(name, null, null, false, cms, site, module);
	}
	/**
	 * Convenience for calling {@link String property(String name, String file, String defaultValue, boolean escapeHtml, CmsJspActionElement cms, Properties site, Properties module)}
	 * - looks at the given file (adhering to <code>&lt;cms:property&gt;</code> file specification), the returned value is not html escaped.
	 * @param name
	 * @param file
	 * @param cms
	 * @param site
	 * @param module
	 * @return
	 */
	public String property(String name, String file, CmsJspActionElement cms, Properties site, Properties module) {
		return property(name, file, null, false, cms, site, module);
	}
	/**
	 * Convenience for calling {@link String property(String name, String file, String defaultValue, boolean escapeHtml, CmsJspActionElement cms, Properties site, Properties module)}
	 * - looks at the given file (adhering to <code>&lt;cms:property&gt;</code> file specification), using defaultValue if property was not
	 * found, the returned value is not html escaped.
	 * 
	 * @param name
	 * @param file
	 * @param defaultValue
	 * @param cms
	 * @param site
	 * @param module
	 * @return
	 */
	
	public String property(String name, String file, String defaultValue, CmsJspActionElement cms, Properties site, Properties module) {
		return property(name, file, defaultValue, false, cms, site, module);
	}
	/**
	 * Mimics the property method on CmsJspActionElement by looking up the property on <code>cms</code>, but in case it is unset
	 * this method will then look for the property in the site setup (if not null) and subsequently in module setup (if not null)
	 * before returning the given default.
	 * 
	 * @param name
	 * @param file
	 * @param defaultValue
	 * @param escapeHtml
	 * @param cms
	 * @param site
	 * @param module
	 * @return
	 */
	public String property(String name, String file, String defaultValue, boolean escapeHtml, CmsJspActionElement cms, Properties site, Properties module) {
		String res = cms.property(name, file, null, false);
		if(res == null && site != null)
			res = site.getProperty(name);
		if(res == null && module != null)
			res = site.getProperty(name);
		if(res == null)
			res = defaultValue;
		if (escapeHtml) {
            res = CmsEncoder.escapeHtml(res);
        }
		return res;
	}

	/**
	 * Returns a resource if it exists (wraps the exception)
	 * @param o
	 * @param filename
	 * @return
	 */
	public static CmsResource exists(CmsObject o, String filename)
	{
	    try {
	    	CmsResource f = o.readResource(filename);
	    	if(f != null)
	    		return f;
	    } catch (CmsException e) {
	    }
	    return null;
	}
	
	/**
	 * @deprecated use CmsObject.existsResource(String resourcename) instead
	 * return true if the resource exists
	 * @param o
	 * @param filename
	 * @return
	 */
	public static boolean fileExists(CmsObject o, String filename) 
	{
	    return o.existsResource(filename);
	}
	
	/**
	 * Returns the CmsResource for the given <code>filename</code>, if it exists and has not been altered after <code>modDate</code>.
	 * @param o
	 * @param filename
	 * @param modDate
	 * @return
	 */
	public static CmsResource existsAndIsUpToDate(CmsObject o, String filename, long modDate)
	{
		try {
			CmsResource f = o.readResource(filename);
			if(f != null && f.getDateLastModified() <= modDate)
				return f;
		} catch (CmsException e) {
		}
		return null;
	}

	/**
	 * Returns the true if the given <code>filename</code>t exists and has not been altered since <code>modDate</code>.
	 * @param o
	 * @param filename
	 * @param modDate
	 * @return
	 */
	public static boolean fileExistsAndIsUpToDate(CmsObject o, String filename, long modDate)
	{
		return (existsAndIsUpToDate(o, filename, modDate) != null);
	}
	
	/**
	 * Returns the CmsResource for the given <code>filename</code>, if it exists and has been altered after <code>modDate</code>.
	 * @param o
	 * @param filename
	 * @param modDate
	 * @return
	 */
	public static CmsResource existsButNeedsUpdate(CmsObject o, String filename, long modDate)
	{
		try {
			CmsResource f = o.readResource(filename);
			if(f != null && f.getDateLastModified() > modDate)
				return f;
		} catch (CmsException e) {
		}
		return null;
	}
	
	/**
	 * Returns true if the given <code>filename</code> exists and has been altered after <code>modDate</code>.
	 * @param o
	 * @param filename
	 * @param modDate
	 * @return
	 */
	public static boolean fileExistsButNeedsUpdate(CmsObject o, String filename, long modDate)
	{
		return (existsButNeedsUpdate(o, filename, modDate) != null);
	}

	/**
	 * 
	 * @param cmso
	 * @return
	 * @deprecated
	 */
	public static boolean isCurrentProjectMutable(CmsObject cmso) {
		return false;
	}
	
	/**
	 * Returns the absolute URL by taking the prefix for the current site as defined by the <code>cms</code> CmsJspBean 
	 * and adding the given <code>uri</code>
	 * <p>Note that this refers to the online site (i.e. it will not return the prefix corresponding to the workplace)
	 * @param cmso
	 * @return
	 */
	public static String absoluteURL(CmsJspBean cms, String uri) 
	{ 
		if(uri == null)
			return absoluteURLPrefix(cms.getCmsObject());
		else
			return absoluteURLPrefix(cms.getCmsObject())+uri;
	}
	
	/**
	 * Returns the absolute URL prefix for the current site as defined by the <code>cms</code> CmsJspBean.
	 * <p>Note that this refers to the online site (i.e. it will not return the prefix corresponding to the workplace)
	 * @param cmso
	 * @return
	 */
	public static String absoluteURLPrefix(CmsJspBean cms) throws MalformedURLException 
	{
		return absoluteURLPrefix(cms.getCmsObject());
	}
	
	/**
	 * Returns the absolute URL prefix for the current site as defined by the <code>cmso</code> CmsObject.
	 * <p>Note that this refers to the online site (i.e. it will not return the prefix corresponding to the workplace)
	 * @param cmso
	 * @return
	 */
	public static String absoluteURLPrefix(CmsObject cmso) {
		//return CmsSiteManager.getCurrentSite(cmso).getUrl();
		return absoluteURLPrefix(CmsSiteManager.getCurrentSite(cmso).getSiteRoot());
	}
	
	/**
	 * Returns the absolute URL prefix for the site defined by the <code>siteRoot</code>.
	 * <p>Note that this refers to the online site (i.e. it will not return the prefix corresponding to the workplace)
	 * @param cmso
	 * @return
	 */
	public static String absoluteURLPrefix(String siteRoot) {
//		 TODO What about OpenCms.getStaticExportManager().getVfsPrefix()?
		return CmsSiteManager.getSite(siteRoot).getUrl();
	}
	/**
	 * Convenience to test if current project is the online project
	 * @param cmso
	 * @return
	 */
	public static boolean isOnlineProject(CmsObject cmso) {
		return cmso.getRequestContext().currentProject().isOnlineProject();
	}
	
	/**
	 * Returns an absolute URL prefix which is usable whether in online or offline project.
	 * 
	 * @param cms
	 * @return
	 */
	public static String siteURLPrefix(CmsJspBean cms) {
		if(isOnlineProject(cms.getCmsObject())) {
			return CmsSiteManager.getCurrentSite(cms.getCmsObject()).getUrl();
		} else {
			HttpServletRequest req = cms.getRequest();
			CmsSiteMatcher matcher = new CmsSiteMatcher(req.getScheme(), req.getServerName(), req.getServerPort());
			return matcher.getUrl();
		}
	}
	
	/**
	 * Used for 1, 2 and tristate buttons. Looking at NavImg, NavImgHover and NavImg properties. Or the NavKey property
	 * which can be used to let the routine auto-generate image file name. If no properties are given a nav key will
	 * be guess by the keyFromUrl method.
	 * Config can hold ekstra values such that ImgPath, ImgExt, ImgHoverSuffix, and ImgSelectSuffix (latter often the same)
	 *
	 * Reserved config keys:
	 * - ${NAV_KEY}:        Read from "NavKey" property, if this is null it is derived from file name
	 * - ${NAV_URL}:        Path to the resource referenced by nav element 
	 * - ${NAV_TEXT}:       Read from "NavText" property - if missing it is set to ""
	 * - ${NAV_DESC}:       Read from "Description" property - if missing it is set to ""
	 * - ${NAV_LEVEL}:      Navigation tree level
	 * - ${LOOP_LEVEL}:     Loop / recursion level
	 * - ${NAV_IMG}:        (only if ImgButtons is set to true)
	 * - ${NAV_IMG_HOVER}:  (only if ImgButtons is set to true)
	 * - ${NAV_IMG_SELECT}: (only if ImgButtons is set to true)
	 */
	public static String buildDDImgNav(CmsJspActionElement cms, String resource, boolean showParent, int topLevel, boolean showCurrent, HashMap config)
	{
	    StringBuffer buffer = new StringBuffer();
	    if(config==null) { // just provide some sort of default.
	    	config = new HashMap(); 
	    	config.put("NavOpen", "<div class=\"Nav\">\n");
	    	config.put("NavClose", "</div>\n");
	    	config.put("NavParent", "<a class=\"NavParent\" href=\"${NAV_URL}\">${NAV_TEXT}</a>\n");
	    	config.put("NavParentDefaultText", "%lt;%lt;");
	    	config.put("NavFolder", "<a class=\"NavFolder\" href=\"${NAV_URL}\">${NAV_TEXT}</a>\n");
	    	config.put("NavFolderSelected", "<a class=\"NavFolderSelected\" href=\"${NAV_URL}\">${NAV_TEXT}</a>\n");
	    	config.put("NavItem", "<a class=\"NavItem\" href=\"${NAV_URL}\">${NAV_TEXT}</a>\n");
	    	config.put("NavItemSelected", "<a class=\"NavItemSelected\" href=\"${NAV_URL}\">${NAV_TEXT}</a>\n");        
	    }
	    
	    HashMap params = new HashMap();
	    String foldername;
	    if(resource == null)
	    	resource = cms.getRequestContext().getUri();
	    foldername = CmsFile.getFolderPath(resource); 
	    String imgPath = (String)config.get("ImgPath"); // This is always prefixed the img url given on the resource
	    if(imgPath==null)
	    	imgPath = ""; 
	    // The next three are used to auto-generate image name if not given per resource
	    String imgExt = (String)config.get("ImgExt");
	    if(imgExt==null)
	    	imgExt = ".gif"; // still most used format...
	    String imgHoverSuffix = (String)config.get("ImgHoverSuffix");
	    if(imgHoverSuffix ==null)
	    	imgHoverSuffix = "_hover";
	    String imgSelectSuffix = (String)config.get("ImgSelectSuffix");
	    if(imgSelectSuffix ==null)
	    	imgSelectSuffix = "_select";
	    
	    // Start to build the navigation - first look at the nav. element for the current page
	    CmsJspNavBuilder builder = cms.getNavigation();
	    CmsJspNavElement currFile = builder.getNavigationForResource(resource);          
	    int level = currFile.getNavTreeLevel();

	    buffer.append(config.get("NavOpen"));

	    // 1: Backlink to overview, but only if parent folder isn't at the site level.
	    if(showParent && level > topLevel) {
	    	String fn = CmsFile.getPathPart(foldername, -2) + "index.html"; ///????
	    	CmsJspNavElement prevFolder = builder.getNavigationForResource(fn);
	    	if (prevFolder.isInNavigation()) {
	    		String key = prevFolder.getProperty("NavKey");
	    		if(key == null)
	    			key = PathUtils.stripExtension(PathUtils.lastPathPart(fn),".");
	    		String img = prevFolder.getProperty("NavImg");
	    		if(img == null)
	    			img = imgPath + key + imgExt;
	    		String imgHover = prevFolder.getProperty("NavImgHover");
	    		if(imgHover == null)
	    			imgHover = imgPath + key + imgHoverSuffix + imgExt;
	    		String imgSelect = prevFolder.getProperty("NavImgSelect");
	    		if(imgSelect == null)
	    			imgSelect = imgPath + key + imgSelectSuffix + imgExt;
	    		params.put("${NAV_KEY}", key);
	    		params.put("${NAV_IMG}", cms.link(img));
	    		params.put("${NAV_IMG_HOVER}", cms.link(imgHover));
	    		params.put("${NAV_IMG_SELECT}", cms.link(imgSelect)); 
	    		params.put("${NAV_URL}", cms.link(fn));
	    		params.put("${NAV_TEXT}", prevFolder.getNavText());
				// deprecated - only here for backward compat.
	    		params.put("${KEY}", key);
	    		params.put("${IMG}", cms.link(img));
	    		params.put("${IMG_HOVER}", cms.link(imgHover));
	    		params.put("${IMG_SELECT}", cms.link(imgSelect)); 
	    		params.put("${HREF}", cms.link(fn));
	    		params.put("${TEXT}", prevFolder.getNavText());
	    		buffer.append(StringUtils.replaceStrings((String)config.get("NavParent"), params));
	    	}
	    } 

	    // 2: Headline for current folder navigation
	    if(showCurrent) {
	    	CmsJspNavElement currFolder = builder.getNavigationForResource(foldername);
	    	if (currFolder.isInNavigation()) {
	    		String key = currFolder.getProperty("NavKey");
	    		if(key == null)
	    			key = PathUtils.stripExtension(PathUtils.lastPathPart(foldername),".");
	    		String img = currFolder.getProperty("NavImg");
	    		if(img == null)
	    			img = imgPath + key + imgExt;
	    		String imgHover = currFolder.getProperty("NavImgHover");
	    		if(imgHover == null)
	    			imgHover = imgPath + key + imgHoverSuffix + imgExt;
	    		String imgSelect = currFolder.getProperty("NavImgSelect");
	    		if(imgSelect == null)
	    			imgSelect = imgPath + key + imgSelectSuffix + imgExt;
	    		params.put("${NAV_KEY}", key);
	    		params.put("${NAV_IMG}", cms.link(img));
	    		params.put("${NAV_IMG_HOVER}", cms.link(imgHover));
	    		params.put("${NAV_IMG_SELECT}", cms.link(imgSelect));
	    		params.put("${NAV_URL}", cms.link(foldername)); 
	    		params.put("${NAV_TEXT}", currFolder.getNavText());
				// deprecated - only here for backward compat.
	    		params.put("${KEY}", key);
	    		params.put("${IMG}", cms.link(img));
	    		params.put("${IMG_HOVER}", cms.link(imgHover));
	    		params.put("${IMG_SELECT}", cms.link(imgSelect));
	    		params.put("${HREF}", cms.link(foldername)); 
	    		params.put("${TEXT}", currFolder.getNavText());
	    		buffer.append(StringUtils.replaceStrings((String)config.get("NavFolderSelected"), params));
	    	}
	    }

	    // 3: List of all pages / subfolders (sorted by NavPos property)
	    List list = builder.getNavigationForFolder(foldername); 
	    Iterator i = list.iterator();
	    while (i.hasNext()) {
	    	CmsJspNavElement ne = (CmsJspNavElement)i.next();
	    	String uri = ne.getResourceName();
	    	String key = ne.getProperty("NavKey");
	    	if(key == null)
	    		key = PathUtils.stripExtension(PathUtils.lastPathPart(uri),".");
	    	String img = ne.getProperty("NavImg");
	    	if(img == null)
	    		img = imgPath + key + imgExt;
	    	String imgHover = ne.getProperty("NavImgHover");
	    	if(imgHover == null)
	    		imgHover = imgPath + key + imgHoverSuffix + imgExt;
	    	String imgSelect = ne.getProperty("NavImgSelect");
	    	if(imgSelect == null)
	    		imgSelect = imgPath + key + imgSelectSuffix + imgExt;
	    	params.put("${NAV_KEY}", key);
	    	params.put("${NAV_IMG}", cms.link(img));
	    	params.put("${NAV_IMG_HOVER}", cms.link(imgHover));
	    	params.put("${NAV_IMG_SELECT}", cms.link(imgSelect));
	    	params.put("${NAV_URL}", cms.link(uri));
	    	params.put("${NAV_TEXT}", ne.getNavText());
			// deprecated - only here for backward compat.
	    	params.put("${KEY}", key);
	    	params.put("${IMG}", cms.link(img));
	    	params.put("${IMG_HOVER}", cms.link(imgHover));
	    	params.put("${IMG_SELECT}", cms.link(imgSelect));
	    	params.put("${HREF}", cms.link(uri));
	    	params.put("${TEXT}", ne.getNavText());
	    	if (!ne.isFolderLink()) {
	    		if (uri.equals(currFile.getResourceName())) {
	    			buffer.append(StringUtils.replaceStrings((String)config.get("NavItemSelected"), params));
	    		} else {
	    			buffer.append(StringUtils.replaceStrings((String)config.get("NavItem"), params));
	    		}
	    	} else {
	    		buffer.append(StringUtils.replaceStrings((String)config.get("NavFolder"), params));
	    	}
	    }

	    buffer.append(config.get("NavClose"));

	    return buffer.toString();
	}

	public static String buildDDNav(CmsJspActionElement cms, String resource, boolean showParent, int topLevel, boolean showCurrent, HashMap config)
	{
		StringBuffer buffer = new StringBuffer();
	    if(config==null) { // just provide some sort of default.
	    	config = new HashMap(); 
	    	config.put("NavOpen", "<div class=\"Nav\">\n");
	    	config.put("NavClose", "</div>\n");
	    	config.put("NavParent", "<a class=\"NavParent\" href=\"${NAV_URL}\">${NAV_TEXT}</a>\n");
	    	config.put("NavParentDefaultText", "%lt;%lt;");
	    	config.put("NavFolder", "<a class=\"NavFolder\" href=\"${NAV_URL}\">${NAV_TEXT}</a>\n");
	    	config.put("NavFolderSelected", "<a class=\"NavFolderSelected\" href=\"${NAV_URL}\">${NAV_TEXT}</a>\n");
	    	config.put("NavItem", "<a class=\"NavItem\" href=\"${HREF}\">${NAV_TEXT}</a>\n");
	    	config.put("NavItemSelected", "<a class=\"NavItemSelected\" href=\"${NAV_URL}\">${NAV_TEXT}</a>\n");        
	    }
	    HashMap params = new HashMap();
	    String foldername;
	    if(resource == null)
	    	resource = cms.getRequestContext().getUri();
	    foldername = CmsFile.getFolderPath(resource);        
	    
	    // Start to build the navigation - first look at the nav. element for the current page
	    CmsJspNavBuilder builder = cms.getNavigation();
	    CmsJspNavElement currFile = builder.getNavigationForResource(resource);          
	    int level = currFile.getNavTreeLevel();

	    buffer.append(config.get("NavOpen"));

	    // 1: Backlink to overview, but only if parent folder isn't at the site level.
	    if(showParent && level > topLevel) {
	    	String fn = CmsFile.getPathPart(foldername, -2) + "index.html";      //????
	    	CmsJspNavElement prevFolder = builder.getNavigationForResource(fn);
	    	params.put("${NAV_URL}", cms.link(fn));
	    	params.put("${HREF}", cms.link(fn));
	    	if (prevFolder.isInNavigation()) {
	    		params.put("${NAV_TEXT}", prevFolder.getNavText());
	    		params.put("${TEXT}", prevFolder.getNavText());
	    	} else {
	    		params.put("${NAV_TEXT}", config.get("NavParentDefaultText"));
	    		params.put("${TEXT}", config.get("NavParentDefaultText"));
	    	}
	    	buffer.append(StringUtils.replaceStrings((String)config.get("NavParent"), params));
	    } 

	    // 2: Headline for current folder navigation
	    if(showCurrent) {
	    	CmsJspNavElement currFolder = builder.getNavigationForResource(foldername);
	    	if (currFolder.isInNavigation()) {
	    		params.put("${NAV_URL}", cms.link(foldername));
	    		params.put("${NAV_TEXT}", currFolder.getNavText());
	    		params.put("${HREF}", cms.link(foldername));
	    		params.put("${TEXT}", currFolder.getNavText());
	    		buffer.append(StringUtils.replaceStrings((String)config.get("NavFolderSelected"), params));
	    	}
	    }

	    // 3: List of all pages / subfolders (sorted by NavPos property)
	    List list = builder.getNavigationForFolder(foldername); 
	    Iterator i = list.iterator();
	    while (i.hasNext()) {
	    	CmsJspNavElement ne = (CmsJspNavElement)i.next();
	    	String uri = ne.getResourceName();
	    	params.put("${NAV_URL}", cms.link(uri));
	    	params.put("${NAV_TEXT}", ne.getNavText());
	    	params.put("${HREF}", cms.link(uri));
	    	params.put("${TEXT}", ne.getNavText());
	    	if (!ne.isFolderLink()) {
	    		if (uri.equals(currFile.getResourceName())) {
	    			buffer.append(StringUtils.replaceStrings((String)config.get("NavItemSelected"), params));
	    		} else {
	    			buffer.append(StringUtils.replaceStrings((String)config.get("NavItem"), params));
	    		}
	    	} else {
	    		buffer.append(StringUtils.replaceStrings((String)config.get("NavFolder"), params));
	    	}
	    }

	    buffer.append(config.get("NavClose"));

	    return buffer.toString();
	}

	// this also "select" folder further upstream than just the parent which is not what we want:
	//static Pattern indexPattern = Pattern.compile("^(.*/)?index\\.[a-zA-Z]+$");
	static Pattern indexPattern = Pattern.compile("^index\\.[a-zA-Z]+$");
	
	/**
	 * Config parameters:
	 * - NavOpen:
	 * - NavClose:
	 * - NavFolder:
	 * - NavFolderSelected:
	 * - NavItem:
	 * - NavItemSelected:
	 * - IndexMatch: if true /foo/bar/ and /foo/bar/index.html are treated as the same url/page which effects 
	 *               when a folder is considered open vs. selected and the index page is selected or not      
	 * - ImgButtons: if true image buttons will be used, it is likely that the following must be defined then
	 * - ImgPath:
	 * - ImgExt:
	 * - ImgHoverSuffix:
	 * - ImgSelectedSuffix:
	 * 
	 * Reserved config keys:
	 * - ${NAV_KEY}:        Read from "NavKey" property, if this is null it is derived from file name
	 * - ${NAV_URL}:        Path to the resource referenced by nav element 
	 * - ${NAV_TEXT}:       Read from "NavText" property - if missing it is set to ""
	 * - ${NAV_DESC}:       Read from "Description" property - if missing it is set to ""
	 * - ${NAV_LEVEL}:      Navigation tree level
	 * - ${LOOP_LEVEL}:     Loop / recursion level
	 * - ${NAV_IMG}:        (only if ImgButtons is set to true)
	 * - ${NAV_IMG_HOVER}:  (only if ImgButtons is set to true)
	 * - ${NAV_IMG_SEL}:    (only if ImgButtons is set to true)
	 * 
	 * @param cms
	 * @param uri - selected uri (needed to highlight appropriate items) - defaults to current uri
	 * @param root - the root of the generated menu tree. If null and firstLevel < 0 then current folder is used
	 *               as root. If root is null and firstLevel >= 0 root is taken to be the firstLevel part of
	 *               the current folder (typically what is needed with sub-trees, then the subroot need not be
	 *               set per folder - assuming that all sub-navs are at the same level though).
	 * @param firstLevel - levels before the firstLevel are not included (set to 0 to include all from the given root)
	 * @param lastLevel - levels after the lastLevel are not included (set to -1 to include all)
	 * @param configs - list of configurations, the loop level defines which one is looked up (defaults to the last one) 
	 * @return
	 */
	public static String buildNavTree(CmsJspActionElement cms, String uri, String root, int firstLevel, int lastLevel, List configs )
	{
		return buildNavTree(cms, uri, root, firstLevel, lastLevel, 1, configs);
	}
	public static String buildNavTree(CmsJspActionElement cms, String uri, String root, int firstLevel, int lastLevel, int unfoldLevel, List configs )
	{
		return buildNavTree(cms, uri, root, firstLevel, lastLevel, unfoldLevel, false, configs);
	}
	public static String buildNavTree(CmsJspActionElement cms, String uri, String root, int firstLevel, int lastLevel, int unfoldLevel, boolean branchOnly, List configs )
	{
	    if(configs==null)
	    	configs = new ArrayList();
	    if(configs.size()==0) {
	    	HashMap config = new HashMap(); 
	    	config.put("NavOpen", "<div class=\"Nav Level${LOOP_LEVEL}\">\n");
	    	config.put("NavClose", "</div>\n");
	    	config.put("NavFolder", "  <a class=\"NavFolder\" href=\"${NAV_URL}\">${NAV_TEXT}</a>\n");
	    	config.put("NavFolderUnfoded", "  <a class=\"NavFolderUnfolded\" href=\"${NAV_URL}\">${NAV_TEXT}</a>\n");
	    	config.put("NavFolderOpen", "  <a class=\"NavFolderOpen\" href=\"${NAV_URL}\">${NAV_TEXT}</a>\n");
	    	config.put("NavFolderSelected", "  <span class=\"NavFolderSelected\">${NAV_TEXT}</span>\n");
	    	config.put("NavItem", "  <a class=\"NavItem\" href=\"${NAV_URL}\">${NAV_TEXT}</a>\n");
	    	config.put("NavItemSelected", "  <span class=\"NavItemSelected\">${NAV_TEXT}</span>\n");
	    	config.put("NavItemDivider", "");
	    	configs.add(config);
	    	config = new HashMap();
	    	config.put("NavOpen", "<div class=\"SubNav Level${LOOP_LEVEL}\">\n");
	    	config.put("NavClose", "</div>\n");
	    	config.put("NavFolder", "  <a class=\"NavFolder\" href=\"${NAV_URL}\">${NAV_TEXT}</a>\n");
	    	config.put("NavFolderUnfolded", "  <a class=\"NavFolderUnfolded\" href=\"${NAV_URL}\">${NAV_TEXT}</a>\n");
	    	config.put("NavFolderOpen", "  <a class=\"NavFolderOpen\" href=\"${NAV_URL}\">${NAV_TEXT}</a>\n");
	    	config.put("NavFolderSelected", "  <span class=\"NavFolderSelected\">${NAV_TEXT}</span>\n");
	    	config.put("NavItem", "  <a class=\"NavItem\" href=\"${NAV_URL}\">${NAV_TEXT}</a>\n");
	    	config.put("NavItemSelected", "  <span class=\"NavItemSelected\">${NAV_TEXT}</span>\n");
	    	config.put("NavItemDivider", "");
	    	configs.add(config);
	    }
		return buildNavTree(cms, uri, root, firstLevel, lastLevel, unfoldLevel, branchOnly, configs, 0, "");
	}
	private static String buildNavTree(CmsJspActionElement cms, String uri, String root, int firstLevel, int lastLevel, int unfoldLevel, boolean branchOnly, List configs , int loopLevel, String indent)
	{
		StringBuffer buffer = new StringBuffer();
	    	    
	    if(uri == null)
	    	uri = cms.getCmsObject().getRequestContext().getUri();
	    String currentFolder = CmsFile.getFolderPath(uri);
		if(root == null) {
			if(firstLevel < 0)
				root = currentFolder;
			else
				root = CmsFile.getPathPart(uri, firstLevel);
		}
//		buffer.append("\n<!-- ROOT="+root+"-- URI="+uri+"-- FIRST="+firstLevel+"-->");
		
	    List navList = cms.getNavigation().getNavigationForFolder(root);
	    if (navList.size() > 0) {
	    	HashMap config;
	    	if(configs.size() > loopLevel)
	    		config = (HashMap)configs.get(loopLevel);
	    	else
	    		config = (HashMap)configs.get(configs.size()-1);
	    	
	    	String imgExt = null, imgPath = null, imgHoverSuffix = null, imgSelectedSuffix = null;
	    	boolean useImgButtons = StringUtils.s4n((String)config.get("ImgButtons")).toLowerCase().equals("true");
	    	if(useImgButtons) {
	    		if((imgPath = (String)config.get("ImgPath")) == null)
	    			imgPath = "";
	    		if((imgExt = (String)config.get("ImgExt")) == null)
	    			imgExt = ".gif";
	    		if((imgHoverSuffix = (String)config.get("ImgHoverSuffix")) == null)
	    			imgHoverSuffix = "_hover";
	    		if((imgSelectedSuffix = (String)config.get("ImgSelectedSuffix")) == null)
	    			imgSelectedSuffix = "_select";
	    	}
			boolean indexMatch = Boolean.valueOf((String)config.get("IndexMatch")).booleanValue();
	    	
		    HashMap parameters = new HashMap();
		    parameters.put("${LOOP_LEVEL}", ""+loopLevel);
		    
		    //buffer.append(indent + StringUtils.replaceStrings((String)config.get("NavOpen"), parameters));

		    Iterator navItem = navList.iterator();
		    boolean isFirst = true;
			boolean isOpened = false;
		    while (navItem.hasNext()) {
		    	CmsJspNavElement nav = (CmsJspNavElement) navItem.next();
		    	String link = nav.getResourceName();
		    	int currentLevel = nav.getNavTreeLevel();
				boolean ignoreBranch = (branchOnly && (!link.startsWith(root) || currentLevel < firstLevel)); 
				if(!ignoreBranch) {
					if(!isOpened && (lastLevel == -1 || currentLevel <= lastLevel)) {
						buffer.append(indent + StringUtils.replaceStrings((String)config.get("NavOpen"), parameters));
						isOpened = true;
					}
					//A_OpenCms.log(I_CmsLogChannels.C_OPENCMS_INFO, "\"" + link + "=" + firstLevel + "<" + currentLevel + "<" + lastLevel + "\"");
					if(currentLevel >= firstLevel && (lastLevel == -1 || currentLevel <= lastLevel)) {
						String key = nav.getProperty("NavKey");
						if(key == null)
							key = PathUtils.keyFromPath(link, firstLevel, lastLevel, true, "_");
						parameters.put("${NAV_KEY}", key);
						parameters.put("${NAV_URL}", cms.link(link));
						parameters.put("${NAV_TEXT}", StringUtils.s4n(nav.getNavText()));
						parameters.put("${NAV_DESC}", StringUtils.s4n(nav.getDescription()));
						parameters.put("${NAV_LEVEL}", ""+currentLevel);
						
						if(useImgButtons) {
							String img = nav.getProperty("NavImg");
							if(img == null)
								img = imgPath + key + imgExt;
							String imgHover = nav.getProperty("NavImgHover");
							if(imgHover == null)
								imgHover = imgPath + key + imgHoverSuffix + imgExt;
							String imgSelect = nav.getProperty("NavImgSelected");
							if(imgSelect == null)
								imgSelect = imgPath + key + imgSelectedSuffix + imgExt;
							parameters.put("${NAV_IMG}", cms.link(img));
							parameters.put("${NAV_IMG_HOVER}", cms.link(imgHover));
							parameters.put("${NAV_IMG_SEL}", cms.link(imgSelect));
						}
						
						if(!isFirst)
							buffer.append((String)config.get("NavItemDivider"));
						else 
							isFirst = false;
						
						if(nav.isFolderLink()) {
							if(uri.equals(link)) {
								buffer.append(indent + StringUtils.replaceStrings((String)config.get("NavFolderSelected"), parameters));
								buffer.append(buildNavTree(cms, uri, link, firstLevel, lastLevel, unfoldLevel, branchOnly, configs, loopLevel+1, indent+"  "));
							}
							else if(uri.startsWith(link)) {
								if(indexMatch && indexPattern.matcher(uri.substring(link.length())).matches()) {
									buffer.append(indent + StringUtils.replaceStrings((String)config.get("NavFolderSelected"), parameters));
									buffer.append(buildNavTree(cms, uri, link, firstLevel, lastLevel, unfoldLevel, branchOnly, configs, loopLevel+1, indent+"  "));									
								} else {
									buffer.append(indent + StringUtils.replaceStrings((String)config.get("NavFolderOpen"), parameters));
									buffer.append(buildNavTree(cms, uri, link, firstLevel, lastLevel, unfoldLevel, branchOnly, configs, loopLevel+1, indent+"  "));
								}
							}
							else if(currentLevel < unfoldLevel) {
								buffer.append(indent + StringUtils.replaceStrings((String)config.get("NavFolderUnfolded"), parameters));
								buffer.append(buildNavTree(cms, uri, link, firstLevel, lastLevel, unfoldLevel, branchOnly, configs, loopLevel+1, indent+"  "));
							}		    				
							else {
								buffer.append(indent + StringUtils.replaceStrings((String)config.get("NavFolder"), parameters));
							}		    				
						} else {
							if(uri.equals(link)) {
								buffer.append(indent + StringUtils.replaceStrings((String)config.get("NavItemSelected"), parameters));
							}
							else if(indexMatch && link.startsWith(uri) && indexPattern.matcher(link.substring(uri.length())).matches()) {
								buffer.append(indent + StringUtils.replaceStrings((String)config.get("NavItemSelected"), parameters));								
							}
							else {
								buffer.append(indent + StringUtils.replaceStrings((String)config.get("NavItem"), parameters));
							}
						}
						parameters.remove("${NAV_KEY}");
						parameters.remove("${NAV_URL}");
						parameters.remove("${NAV_TEXT}");
						parameters.remove("${NAV_DESC}");
						parameters.remove("${NAV_LEVEL}");
						parameters.remove("${LOOP_LEVEL}");
						if(useImgButtons) {
							parameters.remove("${NAV_IMG}");
							parameters.remove("${NAV_IMG_HOVER}");
							parameters.remove("${NAV_IMG_SEL}");
						}
					} else {
						// Dryrun - no output at this level, but traversal should continue
						if(nav.isFolderLink()) {
							buffer.append(buildNavTree(cms, uri, link, firstLevel, lastLevel, unfoldLevel, branchOnly, configs, loopLevel+1, indent+"  "));
						}					
					}
				}
				if(!navItem.hasNext() && isOpened && (lastLevel == -1 || currentLevel <= lastLevel)) {
					buffer.append(indent + StringUtils.replaceStrings((String)config.get("NavClose"), parameters));
					isOpened = true;
				}
		    }
		    
		    //buffer.append(indent + StringUtils.replaceStrings((String)config.get("NavClose"), parameters));
	    }
	    return buffer.toString();
	}

	/**
	 * Config parameters:
	 * - NavOpen:
	 * - NavClose:
	 * - NavFolder:
	 * - NavFolderSelected:
	 * - NavItem:
	 * - NavItemSelected:
	 * - ImgButtons: if true image buttons will be used, it is likely that the following must be defined then
	 * - ImgPath:
	 * - ImgExt:
	 * - ImgHoverSuffix:
	 * - ImgSelectedSuffix:
	 * 
	 * Reserved config keys:
	 * - ${NAV_KEY}:        Read from "NavKey" property, if this is null it is derived from file name
	 * - ${NAV_URL}:        Path to the resource referenced by nav element 
	 * - ${NAV_TEXT}:       Read from "NavText" property - if missing it is set to ""
	 * - ${NAV_DESC}:       Read from "Description" property - if missing it is set to ""
	 * - ${NAV_LEVEL}:      Navigation tree level
	 * - ${LOOP_LEVEL}:     Loop / recursion level
	 * - ${NAV_IMG}:        (only if ImgButtons is set to true)
	 * - ${NAV_IMG_HOVER}:  (only if ImgButtons is set to true)
	 * - ${NAV_IMG_SEL}:    (only if ImgButtons is set to true)
	 * 
	 * @param cms
	 * @param uri - selected uri (needed to highlight appropriate items) - defaults to current uri
	 * @param root - the root of the generated menu tree (if tree level larger than firstLevel this becomes first level)
	 * @param firstLevel - levels before the firstLevel are not included (set to 0 to include all from the given root)
	 * @param lastLevel - levels after the lastLevel are not included (set to -1 to include all)
	 * @param expandLevels - number of menu-levels to show expanded (set to 0 to expand all)
	 * @param configs - list of configurations, the loop level defines which one is looked up (defaults to the last one) 
	 * @return
	 */
	public static String buildNavStack(CmsJspActionElement cms, String uri, String root, int firstLevel, int lastLevel, int expandLevels, List configs )
	{
	    if(configs==null)
	    	configs = new ArrayList();
	    if(configs.size()==0) {
	    	HashMap config = new HashMap(); 
	    	config.put("NavOpen", "<div class=\"Nav Level${LOOP_LEVEL}\">\n");
	    	config.put("NavClose", "</div>\n");
	    	config.put("NavFolder", "  <a class=\"NavFolder\" href=\"${NAV_URL}\">${NAV_TEXT}</a>\n");
	    	config.put("NavFolderOpen", "  <a class=\"NavFolderOpen\" href=\"${NAV_URL}\">${NAV_TEXT}</a>\n");
	    	config.put("NavFolderSelected", "  <span class=\"NavFolderSelected\">${NAV_TEXT}</span>\n");
	    	config.put("NavItem", "  <a class=\"NavItem\" href=\"${NAV_URL}\">${NAV_TEXT}</a>\n");
	    	config.put("NavItemSelected", "  <span class=\"NavItemSelected\">${NAV_TEXT}</span>\n");
	    	config.put("NavItemDivider", "");
	    	configs.add(config);
	    	config = new HashMap();
	    	config.put("NavOpen", "<div class=\"SubNav Level${LOOP_LEVEL}\">\n");
	    	config.put("NavClose", "</div>\n");
	    	config.put("NavFolder", "  <a class=\"NavFolder\" href=\"${NAV_URL}\">${NAV_TEXT}</a>\n");
	    	config.put("NavFolderOpen", "  <a class=\"NavFolderOpen\" href=\"${NAV_URL}\">${NAV_TEXT}</a>\n");
	    	config.put("NavFolderSelected", "  <span class=\"NavFolderSelected\">${NAV_TEXT}</span>\n");
	    	config.put("NavItem", "  <a class=\"NavItem\" href=\"${NAV_URL}\">${NAV_TEXT}</a>\n");
	    	config.put("NavItemSelected", "  <span class=\"NavItemSelected\">${NAV_TEXT}</span>\n");
	    	config.put("NavItemDivider", "");
	    	configs.add(config);
	    }

		StringBuffer buffer = new StringBuffer();
		
	    if(firstLevel < 0)
			firstLevel = 0;
		if(expandLevels < 0)
			expandLevels = 0; // basically a bread crumb then since no levels are expanded
		
	    if(uri == null)
	    	uri = cms.getCmsObject().getRequestContext().getUri();
	    //String currentFolder = CmsFile.getFolderPath(uri);
		CmsJspNavBuilder builder = cms.getNavigation();
		if(root != null) {
			CmsJspNavElement rootNav = builder.getNavigationForResource(root);
			int l = rootNav.getNavTreeLevel();
			if(l > firstLevel)
				firstLevel = l;
		}
	    List navList = builder.getNavigationBreadCrumb(uri, firstLevel, lastLevel, true);
		lastLevel = firstLevel + navList.size() - 1;
		String indent = "  ";
		for(int loopLevel = 0, currentLevel = firstLevel; currentLevel <= lastLevel; loopLevel++, currentLevel++) {
	    	HashMap config;
	    	if(configs.size() > loopLevel)
	    		config = (HashMap)configs.get(loopLevel);
	    	else
	    		config = (HashMap)configs.get(configs.size()-1);
	    	
		    HashMap parameters = new HashMap();
		    parameters.put("${LOOP_LEVEL}", ""+loopLevel);
			
		    buffer.append(StringUtils.replaceStrings((String)config.get("NavOpen"), parameters));

			CmsJspNavElement nav = (CmsJspNavElement)navList.get(loopLevel);
	    	String path = nav.getResourceName();
			String key = nav.getProperty("NavKey");
	    	if(key == null)
				key = PathUtils.keyFromPath(path, firstLevel, lastLevel, true, "_");
	    		//key = PathUtils.stripExtension(PathUtils.lastPathPart(path),".");			    	
    		parameters.put("${NAV_KEY}", key);
    		parameters.put("${NAV_URL}", cms.link(path));
    		parameters.put("${NAV_TEXT}", StringUtils.s4n(nav.getNavText()));
    		parameters.put("${NAV_DESC}", StringUtils.s4n(nav.getDescription()));
    		parameters.put("${NAV_LEVEL}", ""+currentLevel);
			// add the bread crumb
			if(uri.equals(path) || uri.startsWith(path+"index")) {
				buffer.append(StringUtils.replaceStrings((String)config.get("NavFolderSelected"), parameters));
			}
			else if(uri.startsWith(path)) {
				buffer.append(StringUtils.replaceStrings((String)config.get("NavFolderOpen"), parameters));
			}
			else {
				buffer.append(StringUtils.replaceStrings((String)config.get("NavFolder"), parameters));
    		}
			// if we're inside the window we must show the contents of the folder
			if(currentLevel > lastLevel - expandLevels) {
				boolean isFirst = true;
				Iterator folderNavs = builder.getNavigationForFolder(path).iterator();
				while(folderNavs.hasNext()) {
					CmsJspNavElement fnav = (CmsJspNavElement)folderNavs.next();
			    	String link = fnav.getResourceName();
					String k = fnav.getProperty("NavKey");
			    	if(k == null)
						k = PathUtils.keyFromPath(link, firstLevel, lastLevel, true, "_");			    	
		    		parameters.put("${NAV_KEY}", k);
		    		parameters.put("${NAV_URL}", cms.link(link));
		    		parameters.put("${NAV_TEXT}", StringUtils.s4n(fnav.getNavText()));
		    		parameters.put("${NAV_DESC}", StringUtils.s4n(fnav.getDescription()));
					//A_OpenCms.log(I_CmsLogChannels.C_OPENCMS_INFO, "[NavTreeLevel()]  = \"" + link + " - " + currentLevel + "\"");
					if(!isFirst)
						buffer.append((String)config.get("NavItemDivider"));
					else 
						isFirst = false;
		    		if(fnav.isFolderLink()) {
						// selected and open folders should probably be hidden???
		    			if(uri.equals(link) || uri.startsWith(link+"index")) {
		    				buffer.append(indent + StringUtils.replaceStrings((String)config.get("NavFolderSelected"), parameters));
		    			}
		    			else if(uri.startsWith(link)) {
		    				buffer.append(indent + StringUtils.replaceStrings((String)config.get("NavFolderOpen"), parameters));
		    			}
		    			else {
		    				buffer.append(indent + StringUtils.replaceStrings((String)config.get("NavFolder"), parameters));
			    		}		    				
		    		} else {
		    			if(uri.equals(link)) {
		    				buffer.append(indent + StringUtils.replaceStrings((String)config.get("NavItemSelected"), parameters));
		    			} 
		    			else {
		    				buffer.append(indent + StringUtils.replaceStrings((String)config.get("NavItem"), parameters));
		    			}
		    		}
		    	}
			}
		    buffer.append(StringUtils.replaceStrings((String)config.get("NavClose"), parameters));
    		parameters.remove("${NAV_KEY}");
    		parameters.remove("${NAV_URL}");
    		parameters.remove("${NAV_TEXT}");
    		parameters.remove("${NAV_DESC}");
    		parameters.remove("${NAV_LEVEL}");
    		parameters.remove("${LOOP_LEVEL}");
	    }
	    return buffer.toString();
	}
	
	/**
	 * 
	 * @param cms
	 * @param uri - if null current folder is used
	 * @param firstLevel - levels before the firstLevel are not included (set to -1 to include all)
	 * @param lastLevel - levels after the lastLevel are not included (set to -1 to include all)
	 * @param config
	 * @return
	 */
	public static String buildSiteMap(CmsJspActionElement cms, String uri, int firstLevel, int lastLevel, Map config)
	{
		StringBuffer buffer = new StringBuffer();
	    if(config==null) { 
	    	config = new HashMap(); 
	    	config.put("SiteMapOpen", "<ul class=\"SiteMap\">\n");
	    	config.put("SiteMapClose", "</ul>\n");
	    	config.put("SiteMapFolder", "<li class=\"SiteMapFolder\"><a href=\"${NAV_URL}\"><b>${NAV_TEXT}</b></a></li>\n");
	    	config.put("SiteMapItem", "<li class=\"SiteMapItem\"><a href=\"${NAV_URL}\">${NAV_TEXT}</a></li>\n");        
	    }
	    HashMap params = new HashMap();
	    List navList;
	    if(uri != null)
	    	navList = cms.getNavigation().getNavigationForFolder(uri);
	    else
	    	navList = cms.getNavigation().getNavigationForFolder();
	    
	    if (navList.size() > 0) {
		    buffer.append(config.get("SiteMapOpen"));

		    Iterator navItem = navList.iterator();		    
		    while (navItem.hasNext()) {
		    	CmsJspNavElement nav = (CmsJspNavElement) navItem.next();
		    	int currentLevel = nav.getNavTreeLevel();

		    	if(currentLevel >= firstLevel && (lastLevel == -1 || currentLevel <= lastLevel)) {
		    		params.put("${NAV_URL}", cms.link(nav.getResourceName()));
		    		params.put("${NAV_TEXT}", StringUtils.s4n(nav.getNavText()));
		    		params.put("${NAV_DESC}", StringUtils.s4n(nav.getDescription()));
		    		params.put("${NAV_LEVEL}", ""+currentLevel);
		        
		    		if (nav.isFolderLink())
		    			buffer.append(StringUtils.replaceStrings((String)config.get("SiteMapFolder"), params));
		    		else
		    			buffer.append(StringUtils.replaceStrings((String)config.get("SiteMapItem"), params));
		    	}
		        if (nav.isFolderLink())
		        	buffer.append(buildSiteMap(cms, nav.getResourceName(), firstLevel, lastLevel, config));
		    }
		    
		    buffer.append(config.get("SiteMapClose"));
	    }
	    return buffer.toString();
	}
	
	/**
	 * Config parameters:
	 * - BreadCrumbOpen:
	 * - BreadCrumbClose:
	 * - BreadCrumbSep:
	 * - BreadCrumbFolder:
	 * - BreadCrumbItem:
	 * - ImgButtons: if true image buttons will be used, it is likely that the following must be defined then
	 * - ImgPath:
	 * - ImgExt:
	 * - ImgHoverSuffix:
	 * - ImgSelectSuffix:
	 * 
	 * Reserved config keys:
	 * - ${NAV_URL}:        Path to the resource referenced by nav element 
	 * - ${NAV_TEXT}:       Read from "NavText" property - if missing it is set to ""
	 * - ${NAV_DESC}:       Read from "Description" property - if missing it is set to ""
	 * - ${NAV_LEVEL}:      Navigation tree level
	 * 
	 * @param cms
	 * @param uri - uri for which to generate bread crumbs
	 * @param startLevel 
	 * @param endLevel 
	 * @param currentFolder - true to include the current folder
	 * @param currentPage - true to include the current item (page)
	 * @param config
	 * @return
	 */
	public static String buildBreadCrumb(CmsJspActionElement cms, String uri, int startLevel, int endLevel, boolean currentFolder, boolean currentPage, Map config) 
	{
		StringBuffer buffer = new StringBuffer();
		if(config==null) { 
			config = new HashMap(); 
			config.put("BreadCrumbOpen", "<p class=\"BreadCrumb\">");
			config.put("BreadCrumbClose", "</p>\n");
			config.put("BreadCrumbSep", "&nbsp;&gt;&nbsp;");
			config.put("BreadCrumbFolder", "<span class=\"BreadCrumbFolder\"><a href=\"${NAV_URL}\">${NAV_TEXT}</a></span>\n");
			config.put("BreadCrumbItem", "<span class=\"BreadCrumbItem\">${NAV_TEXT}</span>\n");
		}
		HashMap params = new HashMap();
		List navList;
		if(uri != null)
			navList = cms.getNavigation().getNavigationBreadCrumb(uri, startLevel, endLevel, currentFolder);
		else
			navList = cms.getNavigation().getNavigationBreadCrumb(startLevel, currentFolder);
		int currentLevel = 0;
		if (navList.size() > 0 || currentPage) {
			buffer.append(config.get("BreadCrumbOpen"));
			Iterator navItem = navList.iterator();		    
			while (navItem.hasNext()) {
				CmsJspNavElement nav = (CmsJspNavElement) navItem.next();
				currentLevel = nav.getNavTreeLevel();
				String navtext = nav.getNavText();
				if("??? NavText ???".equals(navtext) && nav.getTitle() != null) {
					navtext = nav.getTitle(); // in case the folder has no navtext but need to be displayed anyway
				}
				params.put("${NAV_URL}", cms.link(nav.getResourceName()));
				params.put("${NAV_TEXT}", StringUtils.s4n(navtext));
				params.put("${NAV_DESC}", StringUtils.s4n(nav.getDescription()));
				params.put("${NAV_LEVEL}", ""+currentLevel);
				
				buffer.append(StringUtils.replaceStrings((String)config.get("BreadCrumbFolder"), params));
				
				if(navItem.hasNext())
					buffer.append(config.get("BreadCrumbSep"));
			}
			if(currentPage) {
				if(navList.size() > 0)
					buffer.append(config.get("BreadCrumbSep"));
				String navtext = cms.property("NavText", uri);
				if("??? NavText ???".equals(navtext) && cms.property("Title", uri) != null) {
					navtext = cms.property("Title", uri); // in case the folder has no navtext but need to be displayed anyway
				}
				params.put("${NAV_URL}", cms.link(uri));
				params.put("${NAV_TEXT}", StringUtils.s4n(navtext));
				params.put("${NAV_DESC}", StringUtils.s4n(cms.property("Description", uri)));
				params.put("${NAV_LEVEL}", ""+currentLevel+1);
				buffer.append(StringUtils.replaceStrings((String)config.get("BreadCrumbItem"), params));
			}
			buffer.append(config.get("BreadCrumbClose"));
		}
		return buffer.toString();
	}
	
	//
	// Convenience methods that mimics the corresponding methods on CmsJspActionElement
	//

	public static String label(CmsObject cmso, String label) 
	{
		CmsMessages messages = OpenCms.getWorkplaceManager().getMessages(cmso.getRequestContext().getLocale());
		return messages.key(label);
	}
	
    public static String link(CmsObject cmso, String link) 
	{
        if (link.indexOf(':') >= 0) {
            return OpenCms.getLinkManager().substituteLink(cmso, link);
        } else {
			CmsSite site = CmsSiteManager.getCurrentSite(cmso);
			String abs = CmsLinkManager.getAbsoluteUri(link, site.getUrl()+site.getSiteRoot());
            return OpenCms.getLinkManager().substituteLink(cmso, abs);
        }
    }
	
	public static String property(CmsObject cmso, String property, String action) throws CmsException {
		return property(cmso, property, action, null, false);
	}
	public static String property(CmsObject cmso, String property, String action, String defaultValue) throws CmsException {
		return property(cmso, property, action, defaultValue, false);
	}
	public static String property(CmsObject cmso, String property, String action, String defaultValue, boolean escapeHtml) throws CmsException
	{
		String value;
		
		// if action is not set use default
		if (action == null) {
			action = CmsJspTagProperty.ACTION_VALUES[0];
		}
		
		switch (CmsJspTagProperty.ACTION_VALUES_LIST.indexOf(action)) {
		case 0: // USE_URI
		case 1: // USE_PARENT
			// Read properties of parent (i.e. top requested) file
			value = cmso.readPropertyObject(
					cmso.getRequestContext().getUri(),
					property,
					false).getValue(defaultValue);
			break;
		case 2: // USE_SEARCH
		case 3: // USE_SEARCH_URI
		case 4: // USE_SEARCH_PARENT 
			// Try to find property on parent file and all parent folders
			value = cmso.readPropertyObject(
					cmso.getRequestContext().getUri(),
					property,
					true).getValue(defaultValue);
			break;
		case 5: // USE_ELEMENT_URI
		case 6: // USE_THIS
			// Read properties of this file            
			value = "??? property: \""+action+"\" ???"; //cmso.readPropertyObject(controller.getCurrentRequest().getElementUri(),property,false).getValue(defaultValue);
			break;
		case 7: // USE_SEARCH_ELEMENT_URI
		case 8: // USE_SEARCH_THIS
			// Try to find property on this file and all parent folders
			value = "??? property: \""+action+"\" ???"; //cmso.readPropertyObject(controller.getCurrentRequest().getElementUri(),property,true).getValue(defaultValue);
			break;
		default:
			// Read properties of the file named in the attribute            
			value = cmso.readPropertyObject(
					action,
					property,
					false).getValue(defaultValue);
		}
		if (escapeHtml) {
			value = CmsEncoder.escapeHtml(value);
		}
		return value;
	}
}



